# #################################################
#
#  This file contains a class that creates and  
#    executes the 2D conduction problem
#
#  This includes initial conditions generator,
#    routines to calculate energy, edot, timestep,
#    and to write files.
#
#  Written by Brian O'Shea (oshea@msu.edu)
#
# #################################################

import numpy as np
import matplotlib.pyplot as plt

class HeatConduction2D:
    '''
    HeatConduction2D class: initializes and executes 2D simulations of thermal heat conduction using
    forward-time centered space finite difference methods.

    Inputs:
    
      xvals: tuple of three values that describe the range and grid size in
             the x-direction, [xmin, xmax, Nx].  Default [-10.0, 10.0, 100]

      yvals: tuple of three values that describe the range and grid size in
             the y-direction, [ymin, ymax, Nx].  Default [-10.0, 10.0, 100]

      alpha: coefficient of thermal conductivity.  Default 1.0

      Example call:

         heat = HeatConduction2D(xmin=[-10.0,10.0,100], ymin=[-10.0, 10.0, 100], alpha=1.0)
       
    '''
    
    def __init__(self, xvals = [-10.0, 10.0, 100], yvals = [-10.0, 10.0, 100], alpha=1.0):
        self.xvals = xvals
        self.yvals = yvals
        self.alpha = alpha

        self.dx = (self.xvals[1] - self.xvals[0]) / self.xvals[2]
        self.dy = (self.xvals[1] - self.xvals[0]) / self.xvals[2]


    def create_initial_conditions(self, ic_type=1, background_temp = 1.0, peak_temp = 2.0, number_of_waves = 2, pulse_fwhm = 1.0):
        '''
        Initializes initial conditions (temperature field).

        Inputs:

          ic_type: integer that controls what sort of initial conditions are created.
                   Either 1 (Gaussian pulse) or 2 (sinusoidal variation).  Default 1.

          background_temp: temperature of the background field.  Default 1.0

          peak_temp: peak value of the Gaussian pulse or sinusoidal variation.  Default 2.0

          number_of_waves: number of sine waves in the ic_type=2 conditions.  Default 2

          pulse_fwhm: full width half max (FWHM) of the Gaussian pulse.  Default 1.0

        '''

        # initialize the two fields that we need for this simulation
        self.temperature_field = np.ones( (self.xvals[2]+2, self.yvals[2]+2), dtype='float64' )
        self.edot = np.zeros( (self.xvals[2]+2, self.yvals[2]+2), dtype='float64' )
        
        # arrays that describe the centers of the cells in x,y directions
        self.xpos = np.arange(self.xvals[0]-self.dx/2.0, self.xvals[1]+1.5*self.dx, self.dx, dtype='float64')
        self.ypos = np.arange(self.yvals[0]-self.dy/2.0, self.yvals[1]+1.5*self.dy, self.dy, dtype='float64')

        self.background_temp = background_temp
        self.peak_temp = peak_temp
        
        # initialize gaussian pulse
        if ic_type == 1:
            # set pulse centered at (x,y) = (0.0, 0.0)
            for i in range(self.xpos.size):
                for j in range(self.ypos.size):
                    self.temperature_field[i][j] = background_temp + (peak_temp - background_temp) * \
                        np.exp(-(self.xpos[i]**2 + self.ypos[j]**2)/(2.0*pulse_fwhm**2))

            # set the edges of the box to the background temperature.
            self.temperature_field[0,:] = background_temp
            self.temperature_field[self.xpos.size-1,:] = background_temp
            self.temperature_field[:,0] = background_temp
            self.temperature_field[:,self.ypos.size-1] = background_temp
            
            self.bc_type = ic_type
                   
        # initialize sinusoid
        elif ic_type == 2:
            # sinusoidal pulses in x-direction, with number_of_waves waves.
            k = 2.*np.pi*number_of_waves / (self.xvals[1] - self.xvals[0])
            for i in range(self.xpos.size):
                for j in range(self.ypos.size):
                    self.temperature_field[i][j] = background_temp + 0.5 * (peak_temp - background_temp) * (1.0+np.sin(k*self.xpos[i]))

            # set boundary conditions
            self.temperature_field[0,:] = self.temperature_field[self.xpos.size-2,:]
            self.temperature_field[self.xpos.size-1,:] = self.temperature_field[1,:]
            self.temperature_field[:,0] = self.temperature_field[:,self.ypos.size-2]
            self.temperature_field[:,self.ypos.size-1] = self.temperature_field[:,1]

            self.bc_type = ic_type

        # only two types of ICs.
        else:
            print "IC type", ic_type, "not implemented yet.  Aborting!"
            exit(1)

        return 1
        # set boundary condition type here

    def __set_edge_edot(self):
        '''
        Internal function.  Depending on ICs (and thus boundary conditions) either set edot
        (rate of change of energy) at the boundaries to zero (Gaussian) or periodic (sinusoidal)
        '''

        # gaussian
        if self.bc_type == 1:
            self.edot[0,:] = 0.0
            self.edot[self.xpos.size-1,:] = 0.0
            self.edot[:,0] = 0.0
            self.edot[:,self.ypos.size-1] = 0.0

        # sinusoidal
        elif self.bc_type == 2:
            self.edot[0,:] = self.edot[self.xpos.size-2,:]
            self.edot[self.xpos.size-1,:] = self.edot[1,:]
            self.edot[:,0] = self.edot[:,self.ypos.size-2]
            self.edot[:,self.ypos.size-1] = self.edot[:,1]

        # if something's set incorrectly.
        else:
            print "BC type", self.bc_type, "not implemented yet.  Aborting!"
            exit(1)

        return 1

    def calculate_edot(self):
        '''
        Calculates edot using the temperature field, dx, dy, alpha
        '''
        
        for i in range(1,self.xpos.size-1):
            for j in range(1,self.ypos.size-1):
                self.edot[i][j] = self.alpha * (  (1.0/self.dx**2) * ( self.temperature_field[i+1][j]
                                                                       - 2.0*self.temperature_field[i][j]
                                                                       + self.temperature_field[i-1][j] )
                                                + (1.0/self.dy**2) * ( self.temperature_field[i][j+1]
                                                                       - 2.0*self.temperature_field[i][j]
                                                                       + self.temperature_field[i][j-1] ) )

        self.__set_edge_edot()
        
        return 1

    def calculate_dt(self):
        '''
        Calculates timestep using grid spacing and alpha.
        '''
        
        return 0.3 * 0.5 * self.dx**2 / self.alpha

    def advance_simulation(self):
        '''
        Advances the simulation one step using an internally-calculated timestep.
        Returns: timestep.
        '''

        # calculate time step
        dt = self.calculate_dt()

        # calculate rate of change of temperature
        self.calculate_edot()

        # update temperature field
        self.temperature_field += dt*self.edot

        # return the timestep.
        return dt

    def write_output_file(self, filename="heat_output.png", colormap="jet"):
        '''
        Writes the current state of the temperature field to a file containing a user-supplied file name.

        Inputs:
          filename:  Default "heat_output.png"
          colormap:  Default "jet"

        Note: all colormaps can be found at http://wiki.scipy.org/Cookbook/Matplotlib/Show_colormaps
        
        '''
        
        plt.pcolormesh(self.xpos, self.ypos, self.temperature_field, vmin=self.background_temp, vmax=self.peak_temp,cmap=colormap)

        # Add a colorbar
        plt.colorbar(label="Temperature [K]")
        
        # Tighten up the axis ranges
        plt.xlim(self.xpos.min(), self.xpos.max())
        plt.ylim(self.ypos.min(), self.ypos.max())

        # Save the image
        plt.savefig(filename)
        plt.close()

        return 1
